# -*- coding: utf-8 -*-
'''
agents module for healthcare
============================

This module provides agents in healthcare context.

Classes
-------
WarfarinAgent
    an agent based on Ravvaz et al (2016) paper for Warfarin Dosing

'''
from .warfarin_agent import WarfarinAgent  # noqa: W0611
