# -*- coding: utf-8 -*-
'''
legacy classes for reinforcement learning
=========================================

This module provides classes that are no longer supported.

Classes
-------
Environment:
    The base class that provides minimum required
    functionality for a reinforcement learning environment.

Experiment:
    An environment to explore performance of trained
    agents on subjects.

IterableSubject:
    Now `InstanceGenerator` does the job and goes beyond that!

LegacyReilData

Risk

Snake

WarfarinClusterAgent:
    an agent whose actions are based on clustering of observations.
    Note: This agent will no longer run due to changes in the module
    implementation.
'''
